<?php

namespace Drp\NativeEventApp\GraphQL\Mutations;

use Drp\NativeEventApp\ArrayKeys;
use Drp\NativeEventApp\Models\Device;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Support\Arr;
use Laravel\Passport\ClientRepository;
use Laravel\Passport\Token;
use Laravel\Passport\TokenRepository;
use Lcobucci\JWT\Parser as JwtParser;
use League\OAuth2\Server\AuthorizationServer;
use Psr\Http\Message\ServerRequestInterface;
use Zend\Diactoros\Response as Psr7Response;
use Zend\Diactoros\ServerRequest;

class BaseAuthResolver
{
    /**
     * The authorization server instance.
     *
     * @var AuthorizationServer
     */
    protected $server;

    /**
     * The client repository instance.
     *
     * @var ClientRepository
     */
    protected $clients;

    /**
     * The token repository instance.
     *
     * @var TokenRepository
     */
    protected $tokens;

    /**
     * The JWT token parser instance.
     *
     * @var JwtParser
     */
    protected $jwt;

    public function __construct(
        AuthorizationServer $server,
        ClientRepository $clients,
        TokenRepository $tokens,
        JwtParser $jwt
    ) {
        $this->jwt = $jwt;
        $this->tokens = $tokens;
        $this->server = $server;
        $this->clients = $clients;
    }

    private function buildCredentials(string $grantType, array $args = []): array
    {
        $args = collect($args);
        $credentials = $args->except('directive')->toArray();
        $credentials['client_id'] = config('native-event-app.api.client_id');
        $credentials['client_secret'] = config('native-event-app.api.client_secret');
        $credentials['grant_type'] = $grantType;

        return $credentials;
    }

    protected function buildTokenRequest(array $args, string $grantType): ServerRequestInterface
    {
        $credentials = $this->buildCredentials($grantType, $args);
        return (new ServerRequest())->withParsedBody($credentials);
    }

    protected function getResponse(ServerRequestInterface $request): array
    {
        $response = $this->dispatchRequestToAuthorizationServer($request);

        $this->updateDevice(
            $request,
            $this->saveToken($this->findToken($response['access_token']), $request)
        );

        return ArrayKeys::camelCase($response);
    }

    protected function dispatchRequestToAuthorizationServer(ServerRequestInterface $request): array
    {
        return json_decode($this->server->respondToAccessTokenRequest(
            $request,
            new Psr7Response
        )->getBody()->__toString(), true);
    }

    protected function saveToken(Token $token, ServerRequestInterface $request): Token
    {
        return tap($token, function (Token $token) use ($request): void {
            $this->tokens->save($token->forceFill([
                'user_id' => $this->getUserIdFromRequest($request),
            ]));
        });
    }

    private function getUserIdFromRequest(ServerRequestInterface $request): string
    {
        $body = $request->getParsedBody();
        $authIdentifier = Arr::get($body, 'auth_identifier', null);

        if ($authIdentifier !== null) {
            return $authIdentifier;
        }

        $device = $this->findDevice(Arr::get($body, 'device_id'));

        throw_if($device === null, new ModelNotFoundException);

        return $device->auth_identifier;
    }

    protected function updateDevice(ServerRequestInterface $request, Token $token): void
    {
        $requestData = $request->getParsedBody();

        $device = $this->findDevice($requestData['device_id']);

        if ($device instanceof Device) {
            $device->revokeExistingAccessToken();
            $device->associateAccessToken($token);

            return;
        }

        $this->createDevice($requestData, $token);
    }

    protected function createDevice(array $requestData, Token $token): Device
    {
        return Device::create([
            'id' => $requestData['device_id'],
            'auth_identifier' => $requestData['auth_identifier'],
            'access_token_id' => $token->id,
        ]);
    }

    protected function findDevice(string $deviceId): ?Device
    {
        return Device::find($deviceId);
    }

    protected function findToken(string $token): Token
    {
        return $this->tokens->find(
            $this->jwt->parse($token)->getClaim('jti')
        );
    }
}
