<?php

namespace Drp\NativeEventApp\Models\Concerns;

use Drp\NativeEventApp\Database\TableName;
use Drp\NativeEventApp\Models\Registration;
use Drp\NativeEventApp\Models\Segment;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsToMany;
use Illuminate\Database\Query\Builder as QueryBuilder;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;

trait Segmentable
{
    public function segments(): BelongsToMany
    {
        return $this->belongsToMany(
            Segment::class,
            'nea_segments_' . $this->table
        );
    }

    public function scopeSegmented(Builder $query, Registration $registration): Builder
    {
        $segments = $registration->segments;

        return $query
            ->select()
            ->selectSub(function (QueryBuilder $subQuery) use ($segments) {
                $segmentTable = $this->segments()->getTable();

                return $subQuery->selectRaw("count($segmentTable.id)")
                    ->from($segmentTable)
                    ->where(
                        $this->segments()->getQualifiedForeignPivotKeyName(),
                        '=',
                        DB::raw($this->segments()->getQualifiedParentKeyName())
                    )
                    ->when($segments->isNotEmpty(), function (QueryBuilder $subQuery) use ($segments): QueryBuilder {
                        return $subQuery->whereIn('segment_id', $segments->map->id);
                    });
            }, 'segmentCount')
            ->where(function (Builder $query) use ($segments) {
                return $query->where('segmentCount', '=', $segments->count())
                    ->orWhere('segmentCount', '=', 0);
            });
    }

    public function joinSegments(Collection $segments): void
    {
        $segmentIds = $segments->keyBy->id->map->id;
        $attachedSegmentIds = $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segmentIds)
            ->get()
            ->map(function ($segmentRegistration) {
                return (int) $segmentRegistration->id;
            });

        $this->segments()->attach($segmentIds->except($attachedSegmentIds));
    }

    public function hasAllSegments(Collection $segments): bool
    {
        $segmentIds = $segments->map->id->unique();
        $segmentRegistrations = $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segmentIds)
            ->count();

        return $segmentIds->count() === $segmentRegistrations;
    }

    public function hasAtLeastOneSegment(Collection $segments): bool
    {
        return $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segments->map->id)
            ->count() > 0;
    }
}
