<?php

namespace Drp\NativeEventApp;

use DateInterval;
use Drp\NativeEventApp\Auth\AuthInfoFinder;
use Drp\NativeEventApp\Auth\AuthTokenGrant;
use Drp\NativeEventApp\Auth\CodeGenerator;
use Drp\NativeEventApp\Broadcasters\SocketBroadcaster;
use Drp\NativeEventApp\Contracts\MagicLinkGenerator;
use Drp\NativeEventApp\Contracts\MagicLinkSigner;
use Drp\NativeEventApp\Contracts\QrCode;
use Drp\NativeEventApp\Contracts\UserAgent;
use Drp\NativeEventApp\Enums\AuthMethod;
use Drp\NativeEventApp\Enums\ModuleType;
use Drp\NativeEventApp\Facades\ModuleRegistry;
use Drp\NativeEventApp\GraphQL\ContextFactory;
use Drp\NativeEventApp\Http\ViewComposers\MagicLink;
use Drp\NativeEventApp\Models\Event;
use Drp\NativeEventApp\Models\Page;
use Drp\NativeEventApp\Modules\ModulesUnSerializer;
use Drp\NativeEventApp\Quiz\ApiQuizOperator;
use Drp\NativeEventApp\Quiz\Client\QuizClient;
use Drp\NativeEventApp\Quiz\QuizOperatorManager;
use Drp\NativeEventApp\Templates\DataProviders\DataProviderRegistry;
use Faker\Factory as FakerFactory;
use Faker\Generator;
use GuzzleHttp\Client;
use Illuminate\Container\Container;
use Illuminate\Database\Eloquent\Factory;
use Illuminate\Database\Eloquent\Relations\Relation;
use Illuminate\Support\Facades\View;
use Illuminate\Support\ServiceProvider as IlluminateServiceProvider;
use Jenssegers\Agent\Agent;
use Laravel\Passport\Bridge\AuthCodeRepository;
use Laravel\Passport\Bridge\RefreshTokenRepository;
use Laravel\Passport\Passport;
use League\OAuth2\Server\AuthorizationServer;
use Nuwave\Lighthouse\Schema\TypeRegistry;
use Nuwave\Lighthouse\Schema\Types\LaravelEnumType;
use Nuwave\Lighthouse\Subscriptions\BroadcastManager;
use Nuwave\Lighthouse\Support\Contracts\CreatesContext;
use SimpleSoftwareIO\QrCode\BaconQrCodeGenerator;

/**
 * @SuppressWarnings(PHPMD.CouplingBetweenObjects)
 */
class ServiceProvider extends IlluminateServiceProvider
{
    public function register(): void
    {
        $configPath = __DIR__ . '/../config/native-event-app.php';
        $this->mergeConfigFrom($configPath, 'native-event-app');
        $this->loadViewsFrom(__DIR__ . '/Http/views', 'NativeEventApp');
        $this->loadMigrationsFrom(__DIR__ . '/../database/migrations');

        View::composer('NativeEventApp::magic', MagicLink::class);

        $this->app->singleton(AuthInfoFinder::class);
        $this->app->singleton(CreatesContext::class, ContextFactory::class);

        $this->registerEloquentFactory();
        $this->registerQuizOperator();

        Relation::morphMap([
            'events' => Event::class,
            'pages' => Page::class,
        ]);

        $this->app->singleton(ModulesUnSerializer::class, static function (Container $app) {
            return new ModulesUnSerializer($app);
        });

        $this->app->singleton(DataProviderRegistry::class, function () {
            return new DataProviderRegistry();
        });
    }

    public function boot(TypeRegistry $typeRegistry): void
    {
        $this->loadTranslationsFrom(__DIR__ . '/../lang', 'NativeEventApp');

        app(BroadcastManager::class)->extend('socket', function ($app) {
            return $app->make(SocketBroadcaster::class);
        });

        $this->registerOauthServer();
        $this->registerEnums($typeRegistry);
        $this->registerMagicLink();

        if ($this->app->runningInConsole() === false) {
            return;
        }

        $this->registerPublishes();
    }

    private function registerEnums(TypeRegistry $typeRegistry): void
    {
        $typeRegistry
            ->register(
                new LaravelEnumType(AuthMethod::class)
            )
            ->register(
                new ModuleType(ModuleRegistry::modulesForEnums())
            );
    }

    private function registerEloquentFactory(): void
    {
        $this->app->singleton(Generator::class, function ($app) {
            return FakerFactory::create($app['config']->get('app.faker_locale', 'en_GB'));
        });

        $this->app->singleton('SeededFaker', function ($app) {
            $faker = $app->make(Generator::class);
            $faker->seed('NEA');

            return $faker;
        });

        $this->app->singleton(Factory::class, function ($app) {
            return Factory::construct(
                $app['SeededFaker'],
                database_path('factories')
            );
        });
    }

    private function registerOauthServer(): void
    {
        /** @var AuthorizationServer $server */
        $server = $this->app->get(AuthorizationServer::class);
        $tokenExpiry = new DateInterval(config('native-event-app.access_token.ttl'));
        $refreshTokenExpiry = new DateInterval('P1M');

        $this->app->bind(AuthTokenGrant::class, function () use ($refreshTokenExpiry, $tokenExpiry) {
            return new AuthTokenGrant(
                $this->app->make(AuthCodeRepository::class),
                $this->app->make(RefreshTokenRepository::class),
                $tokenExpiry,
                $refreshTokenExpiry
            );
        });

        $server->enableGrantType($this->makeAuthTokenGrant(), $tokenExpiry);

        Passport::tokensCan([
            'all' => 'All',
        ]);

        Passport::setDefaultScope([
            'all',
        ]);
    }

    private function makeAuthTokenGrant(): AuthTokenGrant
    {
        return tap($this->app->make(AuthTokenGrant::class), function ($grant): void {
            $grant->setRefreshTokenTTL(Passport::refreshTokensExpireIn());
        });
    }

    private function registerMagicLink(): void
    {
        $this->app->bind(QrCode::class, function ($app) {
            return new Auth\QrCode($app->make(BaconQrCodeGenerator::class));
        });
        $this->app->bind(UserAgent::class, function ($app) {
            return new Auth\UserAgent($app->make(Agent::class));
        });

        $this->app->bind(MagicLinkGenerator::class, function ($app) {
            return new Auth\MagicLinkGenerator(
                $app->make(MagicLinkSigner::class),
                $app->make(QrCode::class),
                url(config('lighthouse.route.uri')),
                new DateInterval(config('native-event-app.qr_code.ttl'))
            );
        });

        $this->app->bind(MagicLinkSigner::class, function () {
            return new Auth\MagicLinkSigner(config('app.key'));
        });

        $this->app->bind(AuthInfoFinder::class, function ($app) {
            return new Auth\AuthInfoFinder(
                $app->make(MagicLinkSigner::class)
            );
        });

        $this->app->bind(CodeGenerator::class, function () {
            return new CodeGenerator('ABCDEFGHJKLMNPQRSTUVWXYZ234567');
        });
    }

    private function registerQuizOperator(): void
    {
        $this->app->bind(QuizOperatorManager::class, function ($app) {
            return new QuizOperatorManager($app);
        });

        $this->app->bind(ApiQuizOperator::class, function () {
            return new ApiQuizOperator(
                new QuizClient(
                    new Client(['base_uri' => config('native-event-app.socket_server.host')])
                )
            );
        });
    }

    private function registerPublishes(): void
    {
        $this->publishes([
            __DIR__ . '/../config/native-event-app.php' => $this->getConfigPath(),
        ], 'config');

        $this->publishes([
            __DIR__ . '/../database/default-migrations' => database_path('migrations'),
        ], 'migrations');
    }

    private function getConfigPath(): string
    {
        return config_path('native-event-app.php');
    }
}
