<?php

namespace Drp\NativeEventApp\Test\Api\Auth;

use Drp\NativeEventApp\Auth\Facades\AuthInfoFinder;
use Drp\NativeEventApp\Auth\Facades\CodeGenerator;
use Drp\NativeEventApp\Notifications\SendAuthCode;
use Drp\NativeEventApp\Test\Api\GraphQLTestCase;
use Drp\NativeEventApp\Test\Support\Fakes\Registration;
use Illuminate\Contracts\Mail\Mailer;
use Illuminate\Support\Facades\Notification;
use Nexmo\Client as NexmoClient;
use Spatie\Snapshots\MatchesSnapshots;

class AuthRequestTest extends GraphQLTestCase
{
    use MatchesSnapshots;

    /** @test */
    public function email_auth_token_to_auth_source(): void
    {
        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');
        Notification::fake();

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:EMAIL)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));

        Notification::assertSentTo(
            $registration,
            SendAuthCode::class,
            function ($notification, $channels) {
                $this->assertEquals(['mail'], $channels);
                $this->assertEquals('CODE', $notification->tokenCode);
                return true;
            }
        );
    }

    /** @test */
    public function sms_auth_token_to_auth_source(): void
    {
        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');
        Notification::fake();

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:SMS)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));

        Notification::assertSentTo(
            $registration,
            SendAuthCode::class,
            function ($notification, $channels) {
                $this->assertEquals(['nexmo'], $channels);
                $this->assertEquals('CODE', $notification->tokenCode);
                return true;
            }
        );
    }

    /** @test */
    public function handles_sms_sending_issues(): void
    {
        $this->mock(NexmoClient::class, static function ($mock): void {
            $mock
                ->shouldReceive('message')
                ->andThrow(new \Nexmo\Client\Exception\Exception('Original error', 400));
        });

        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:SMS)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }

    /** @test */
    public function handles_invalid_email(): void
    {
        $this->mock(Mailer::class, static function ($mock): void {
            $mock
                ->shouldReceive('send')
                ->andThrow(new \Swift_RfcComplianceException('Invalid email'));
        });

        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:EMAIL)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }

    /** @test */
    public function handles_email_sending_issues(): void
    {
        $this->mock(Mailer::class, static function ($mock): void {
            $mock
                ->shouldReceive('send')
                ->andThrow(new \Swift_SwiftException('Sending error'));
        });

        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:EMAIL)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }

    /** @test */
    public function handles_sending_issues(): void
    {
        $this->mock(Mailer::class, static function ($mock): void {
            $mock
                ->shouldReceive('send')
                ->andThrow(new \Exception('Some other error'));
        });

        $registration = new Registration;
        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->andReturn('CODE');

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:EMAIL)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }

    /** @test */
    public function validates_sms_auth_method(): void
    {
        $registration = new Registration;
        $registration->phone = '11111111';

        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->never();

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:SMS)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }

    /** @test */
    public function validates_email_auth_method(): void
    {
        $registration = new Registration;
        $registration->email = 'invalidemail.com';

        AuthInfoFinder::shouldReceive('get')->andReturn($registration);
        CodeGenerator::shouldReceive('authRequestCode')->never();

        $query = 'mutation {
        authRequest(authIdentifier:"1", expiry:"1234", signature: "1234", authMethod:EMAIL)
        }';

        $this->assertMatchesJsonSnapshot($this->graphQlRequest($query));
    }
}
